@@ -1,7 +1,7 @@
 """
 forms for django-form-utils
 
-Time-stamp: <2009-11-25 01:30:16 carljm forms.py>
+Time-stamp: <2010-04-28 02:57:16 carljm forms.py>
 
 """
 from copy import deepcopy
@@ -19,7 +19,7 @@ class Fieldset(object):
         self.form = form
         self.boundfields = boundfields
         if legend is None: legend = name
-        self.legend = mark_safe(legend)
+        self.legend = legend and mark_safe(legend)
         self.classes = classes
         self.description = mark_safe(description)
         self.name = name
@@ -43,11 +43,26 @@ class FieldsetCollection(object):
     def __init__(self, form, fieldsets):
         self.form = form
         self.fieldsets = fieldsets
+        self._cached_fieldsets = []
 
     def __len__(self):
         return len(self.fieldsets) or 1
 
     def __iter__(self):
+        if not self._cached_fieldsets:
+            self._gather_fieldsets()
+        for field in self._cached_fieldsets:
+            yield field
+
+    def __getitem__(self, key):
+        if not self._cached_fieldsets:
+            self._gather_fieldsets()
+        for field in self._cached_fieldsets:
+            if field.name == key:
+                return field
+        raise KeyError
+
+    def _gather_fieldsets(self):
         if not self.fieldsets:
             self.fieldsets = (('main', {'fields': self.form.fields.keys(),
                                         'legend': ''}),)
@@ -59,10 +74,10 @@ class FieldsetCollection(object):
                 raise ValueError("Fieldset definition must include 'fields' option." )
             boundfields = [forms.forms.BoundField(self.form, self.form.fields[n], n)
                            for n in field_names]
-            yield Fieldset(self.form, name, boundfields,
-                           options.get('legend', None),
-                           ' '.join(options.get('classes', ())),
-                           options.get('description', ''))
+            self._cached_fieldsets.append(Fieldset(self.form, name,
+                boundfields, options.get('legend', None), 
+                ' '.join(options.get('classes', ())),
+                options.get('description', '')))
 
 def _get_meta_attr(attrs, attr, default):
     try:
@@ -71,6 +86,13 @@ def _get_meta_attr(attrs, attr, default):
         ret = default
     return ret
             
+def _set_meta_attr(attrs, attr, value):
+    try:
+        setattr(attrs['Meta'], attr, value)
+        return True
+    except KeyError:
+        return False
+            
 def get_fieldsets(bases, attrs):
     """
     Get the fieldsets definition from the inner Meta class.
@@ -83,15 +105,24 @@ def get_fieldsets(bases, attrs):
             fieldsets = getattr(base, 'base_fieldsets', None)
             if fieldsets is not None:
                 break
-    fieldsets = fieldsets or ()
-    # try looping through fieldsets to catch malformed setting early
-    try:
-        for name, option in fieldsets:
-            pass
-    except TypeError:
-        raise ValueError('"fieldsets" must be an iterable of two-tuples')
+    fieldsets = fieldsets or []
     return fieldsets
 
+def get_fields_from_fieldsets(fieldsets):
+    """
+    Get a list of all fields included in a fieldsets definition.
+
+    """
+    fields = []
+    try:
+        for name, options in fieldsets:
+            fields.extend(options['fields'])
+    except (TypeError, KeyError):
+        raise ValueError('"fieldsets" must be an iterable of two-tuples, '
+                         'and the second tuple must be a dictionary '
+                         'with a "fields" key')
+    return fields
+
 def get_row_attrs(bases, attrs):
     """
     Get the row_attrs definition from the inner Meta class.
@@ -115,6 +146,10 @@ def _mark_row_attrs(bf, form):
 class BetterFormBaseMetaclass(type):
     def __new__(cls, name, bases, attrs):
         attrs['base_fieldsets'] = get_fieldsets(bases, attrs)
+        fields = get_fields_from_fieldsets(attrs['base_fieldsets'])
+        if (_get_meta_attr(attrs, 'fields', None) is None and
+            _get_meta_attr(attrs, 'exclude', None) is None):
+            _set_meta_attr(attrs, 'fields', fields)
         attrs['base_row_attrs'] = get_row_attrs(bases, attrs)
         new_class = super(BetterFormBaseMetaclass,
                           cls).__new__(cls, name, bases, attrs)
@@ -183,11 +218,15 @@ class BetterBaseForm(object):
     def __init__(self, *args, **kwargs):
         self._fieldsets = deepcopy(self.base_fieldsets)
         self._row_attrs = deepcopy(self.base_row_attrs)
+        self._fieldset_collection = None
         super(BetterBaseForm, self).__init__(*args, **kwargs)
 
     @property
     def fieldsets(self):
-        return FieldsetCollection(self, self._fieldsets)
+        if not self._fieldset_collection:
+            self._fieldset_collection = FieldsetCollection(self,
+                                                           self._fieldsets)
+        return self._fieldset_collection
 
     def __iter__(self):
         for bf in super(BetterBaseForm, self).__iter__():
